#ifndef SV_HELPER_HPP_INCLUDED__
#define SV_HELPER_HPP_INCLUDED__

#include <type_traits>
#include <stdexcept>
#include <string_view>

namespace sv
{
  struct tail {}; /**< Empty type used to indicate string end semantics. */

  /** \brief Make a default-constructed view.
   * \param -none-
   * \return empty view */
  template<typename charT, typename traitsT = typename ::std::char_traits<charT>>
  constexpr auto make() noexcept
  {
    return ::std::basic_string_view<charT, traitsT>{};
  }

  /** \brief Make a view of the first `count` characters.
   * \param ptr    pointer to array of characters
   * \param count  number of characters to be used
   * \return view of the first `count` characters
   * NOTE: validity of the range can't be evaluated by the function */
  template<typename charT, typename traitsT = typename ::std::char_traits<charT>>
  constexpr auto make(const charT* ptr, const size_t count)
  {
    if (ptr == nullptr && count != 0)
      throw ::std::invalid_argument{ "sv::make: ptr == nullptr && count != 0" };

    return ::std::basic_string_view<charT, traitsT>{ ptr, count };
  }

  /** \brief Make a view of `count` characters with `off` characters skipped from the beginning.
   * \param off    number of characters to be skipped from the beginning of the string
   * \param ptr    pointer to array of characters
   * \param count  number of characters to be used
   * \return view of range [off, off + count)
   * NOTE: validity of the range can't be evaluated by the function */
  template<typename charT>
  constexpr auto make(const size_t off, const charT* ptr, const size_t count)
  {
    return make(ptr, off + count).substr(off, count);
  }

  /** \brief Make a view of a null-terminated character string.
   * \param ntstr  pointer to a null-terminated string of characters
   * \return view of the character string
   * NOTE: null-termination is mandatory, but doesn't count in size() */
  template<typename charT, typename traitsT = typename ::std::char_traits<charT>>
  constexpr auto make(const charT* ntstr)
  {
    if (ntstr == nullptr)
      throw ::std::invalid_argument{ "sv::make: ntstr == nullptr" };

    return ::std::basic_string_view<charT, traitsT>{ ntstr };
  }

  /** \brief Make a view of a null-terminated character string with `off` characters skipped from the beginning.
   * \param off    number of characters to be skipped from the beginning of the string
   * \param ntstr  pointer to a null-terminated string of characters
   * \return view of range [off, string end)
   * NOTE: null-termination is mandatory, but doesn't count in size() */
  template<typename charT>
  constexpr auto make(const size_t off, const charT* ntstr)
  {
    return make(ntstr).substr(off);
  }

  /** \brief Make a view of the last `count` characters of a null-terminated character string.
   * \param ntstr  pointer to a null-terminated string of characters
   * \param count  number of characters to be used from the end of the string
   * \param tail   object of type sv::tail which indicates string end semantics
   * \return view of range [string end - count, string end)
   * NOTE: null-termination is mandatory, but doesn't count in size() */
  template<typename charT>
  constexpr auto make(const charT* ntstr, size_t count, tail&&)
  {
    const auto sv{ make(ntstr) };
    count = count <= sv.size() ? count : sv.size();
    return sv.substr(sv.size() - count, count);
  }

  /** \brief Make a copy of another view.
   * \param sv  referenced view
   * \return copy of sv */
  template<typename charT, typename traitsT>
  constexpr auto make(const ::std::basic_string_view<charT, traitsT>& sv) noexcept
  {
    return ::std::basic_string_view<charT, traitsT>{ sv };
  }

  /** \brief Make a view of the first `count` characters of another view.
   * \param sv     referenced view
   * \param count  number of characters to be used
   * \return view of the first `count` characters */
  template<typename charT, typename traitsT>
  constexpr auto make(const ::std::basic_string_view<charT, traitsT>& sv, const size_t count)
  {
    return sv.substr(0, count);
  }

  /** \brief Make a view of `count` characters of another view with `off` characters skipped from the beginning.
   * \param off    number of characters to be skipped from the beginning of the buffer
   * \param sv     referenced view
   * \param count  number of characters to be used
   * \return view of range [off, off + count) */
  template<typename charT, typename traitsT>
  constexpr auto make(const size_t off, const ::std::basic_string_view<charT, traitsT>& sv, const size_t count)
  {
    return sv.substr(off, count);
  }

  /** \brief Make a view of another view with `off` characters skipped from the beginning.
   * \param off  number of characters to be skipped from the beginning of the buffer
   * \param sv   referenced view
   * \return view of range [off, string end) */
  template<typename charT, typename traitsT>
  constexpr auto make(const size_t off, const ::std::basic_string_view<charT, traitsT>& sv)
  {
    return sv.substr(off);
  }

  /** \brief Make a view of the last `count` characters of another view.
   * \param sv     referenced view
   * \param count  number of characters to be used from the end of the string
   * \param tail   object of type sv::tail which indicates string end semantics
   * \return view of range [string end - count, string end) */
  template<typename charT, typename traitsT>
  constexpr auto make(const ::std::basic_string_view<charT, traitsT>& sv, size_t count, tail&&)
  {
    count = count <= sv.size() ? count : sv.size();
    return sv.substr(sv.size() - count, count);
  }

  /** \brief Make a view of a string-like container buffer.
     * \param string_like  referenced object with a contiguous buffer containing a string of characters
     * \return view of the buffer
   * NOTE: std::basic_string<charT>, std::array<charT, N> and std::vector<charT> are string-like if a std::basic_string_view<charT> is constructible */
  template<typename classT, typename charT = typename classT::value_type, typename traitsT = typename ::std::char_traits<charT>, typename sizeretT = decltype(::std::declval<classT&>().size()), typename dataretT = decltype(::std::declval<classT&>().data())>
  constexpr auto make(const classT& string_like) noexcept
  {
    return ::std::basic_string_view<charT, traitsT>{ string_like.data(), string_like.size() };
  }

  /** \brief Make a view of the first `count` characters of a string-like container buffer.
   * \param string_like  referenced object with a contiguous buffer containing a string of characters
   * \param count        number of characters to be used
   * \return view of the first `count` characters
   * NOTE: std::basic_string<charT>, std::array<charT, N> and std::vector<charT> are string-like if a std::basic_string_view<charT> is constructible */
  template<typename classT, typename charT = typename classT::value_type, typename traitsT = typename ::std::char_traits<charT>, typename sizeretT = decltype(::std::declval<classT&>().size()), typename dataretT = decltype(::std::declval<classT&>().data())>
  constexpr auto make(const classT& string_like, const size_t count)
  {
    return make(string_like).substr(0, count);
  }

  /** \brief Make a view of `count` characters of a string-like container buffer with `off` characters skipped from the beginning.
   * \param off          number of characters to be skipped from the beginning of the buffer
   * \param string_like  referenced object with a contiguous buffer containing a string of characters
   * \param count        number of characters to be used
   * \return view of range [off, off + count)
   * NOTE: std::basic_string<charT>, std::array<charT, N> and std::vector<charT> are string-like if a std::basic_string_view<charT> is constructible */
  template<typename classT, typename charT = typename classT::value_type, typename traitsT = typename ::std::char_traits<charT>, typename sizeretT = decltype(::std::declval<classT&>().size()), typename dataretT = decltype(::std::declval<classT&>().data())>
  constexpr auto make(const size_t off, const classT& string_like, const size_t count)
  {
    return make(string_like).substr(off, count);
  }

  /** \brief Make a view of a of a string-like container buffer with `off` characters skipped from the beginning.
   * \param off          number of characters to be skipped from the beginning of the buffer
   * \param string_like  referenced object with a contiguous buffer containing a string of characters
   * \return view of range [off, string end)
   * NOTE: std::basic_string<charT>, std::array<charT, N> and std::vector<charT> are string-like if a std::basic_string_view<charT> is constructible */
  template<typename classT, typename charT = typename classT::value_type, typename traitsT = typename ::std::char_traits<charT>, typename sizeretT = decltype(::std::declval<classT&>().size()), typename dataretT = decltype(::std::declval<classT&>().data())>
  constexpr auto make(const size_t off, const classT& string_like)
  {
    return make(string_like).substr(off);
  }

  /** \brief Make a view of the last `count` characters of a string-like container buffer.
   * \param string_like  referenced object with a contiguous buffer containing a string of characters
   * \param count        number of characters to be used from the end of the string
   * \param tail         object of type sv::tail which indicates string end semantics
   * \return view of range [string end - count, string end)
   * NOTE: std::basic_string<charT>, std::array<charT, N> and std::vector<charT> are string-like if a std::basic_string_view<charT> is constructible */
  template<typename classT, typename charT = typename classT::value_type, typename traitsT = typename ::std::char_traits<charT>, typename sizeretT = decltype(::std::declval<classT&>().size()), typename dataretT = decltype(::std::declval<classT&>().data())>
  constexpr auto make(const classT& string_like, size_t count, tail&&)
  {
    count = count <= string_like.size() ? count : string_like.size();
    return make(string_like).substr(string_like.size() - count, count);
  }
}

#endif // SV_HELPER_HPP_INCLUDED__
